import multiprocessing as mp
import os
import pytest
import random

import pandas as pd

import tuneconfig
from tuneconfig.experiment import Experiment

import os, sys
sys.path.insert(0, os.path.abspath("tests"))

import conftest


def test_run(experiment):
    num_samples = num_workers = 5
    results = experiment.run(conftest.exec_func, num_samples, num_workers, verbose=True)
    assert len(results) == len(experiment.config_iterator)
    for trial_id, trial_results in results.items():
        assert len(trial_results) == num_samples
        for pname, config, run_dir, run_id in trial_results:
            assert run_dir not in config
            assert os.path.exists(run_dir)



import os
import random
import shutil

import multiprocessing as mp
import pandas as pd
import pytest

import tuneconfig
import tuneconfig.experiment
import tuneconfig.analysis


def exec_func(config):
    logdir = config["logdir"]
    run_id = logdir.split("/")[-1]
    del config["logdir"]
    os.makedirs(logdir)

    df = pd.DataFrame({
        "foo": pd.Series([random.gauss(10.0, 1.0) for _ in range(10)]),
        "baz": pd.Series([random.gauss(20.0, 2.0) for _ in range(10)]),
        "bar": pd.Series([random.gauss(30.0, 3.0) for _ in range(10)]),
    })
    filepath = os.path.join(logdir, "data.csv")
    df.to_csv(filepath, index=False)

    df = pd.DataFrame({
        "test": pd.Series([random.randrange(-10.0, 10.0) for _ in range(20)]),
    })
    filepath = os.path.join(logdir, "metric.csv")
    df.to_csv(filepath, index=False)

    return (mp.current_process().pid, config, logdir, run_id)


@pytest.fixture(scope="session")
def config_factory():
    def format_fn(param):
        fmt = {
            "batch_size": "batch",
            "horizon": "hr",
            "learning_rate": "lr",
            "optimizer": "opt",
            "num_samples": None
        }
        return fmt.get(param, param)

    return tuneconfig.TuneConfig({
        "batch_size": tuneconfig.grid_search([32, 64, 128]),
        "horizon": 40,
        "learning_rate": tuneconfig.grid_search([0.01, 0.1]),
        "epochs": 1000,
        "optimizer": tuneconfig.grid_search(["Adam", "RMSProp", "GradientDescent"]),
        "num_samples": 10
        },
        format_fn=format_fn,
    )


@pytest.fixture(scope="session")
def experiment(config_factory):
    logdir = "/tmp/tuneconfig"
    experiment = tuneconfig.experiment.Experiment(config_factory, logdir)
    experiment.start()
    yield experiment
    shutil.rmtree(logdir)


@pytest.fixture(scope="session")
def analysis(experiment):
    num_samples = num_workers = 10
    _ = experiment.run(exec_func, num_samples, num_workers)
    analysis = tuneconfig.analysis.ExperimentAnalysis(experiment.logdir)
    analysis.setup()
    return analysis


@pytest.fixture(scope="session")
def trial(analysis):
    return analysis[0]



import itertools
import json
import os


def grid_search(lst):
    return ParamsIterator(lst)


class ParamsIterator:

    def __init__(self, lst):
        self._lst = lst

    def __iter__(self):
        return iter(self._lst)


class TuneConfig:

    def __init__(self, config_dict, format_fn=None):
        self._config_dict = config_dict
        self._format_fn = format_fn

        self._base_dict = {}
        self._params_iterators = {}

        for param, value in self._config_dict.items():
            if isinstance(value, ParamsIterator):
                self._params_iterators[param] = list(value)
            else:
                self._base_dict[param] = value

        self._params = list(self._params_iterators.keys())
        self._values = list(self._params_iterators.values())
        self._value_instantiations = list(itertools.product(*self._values))

    def __len__(self):
        return len(self._value_instantiations)

    def __getitem__(self, i):
        values = self._value_instantiations[i]
        return {
            **self._base_dict,
            **dict(zip(self._params, values))
        }

    def __iter__(self):
        self._index = 0
        return self

    def __next__(self):
        i = self._index
        self._index += 1

        if i == len(self._value_instantiations):
            raise StopIteration

        return self[i]

    def _trial_id(self, config):
        assignments = []

        for param in sorted(config):
            fmt_param = param
            if self._format_fn:
                fmt_param = self._format_fn(param)

            if fmt_param is None:
                continue

            value = config[param]
            assignments.append(f"{fmt_param}={value}")

        return '/'.join(assignments)

    def dump(self, basepath, ignore=None):
        json_files_created = []

        for config in self:
            if not self._is_config_valid(config, ignore):
                continue

            dirpath = os.path.join(basepath, self._trial_id(config))

            if not os.path.exists(dirpath):
                os.makedirs(dirpath)

            filename = "config.json"
            filepath = os.path.join(dirpath, filename)

            with open(filepath, "w") as file:
                json.dump(config, file)
                json_files_created.append(filepath)

        return json_files_created

    def _is_config_valid(self, config, ignore):
        if ignore is None:
            return True

        if isinstance(ignore, dict):
            ignore = [ignore]

        valid = True
        for no_config in ignore:
            if all(config.get(key) == value for key, value in no_config.items()):
                valid = False
                break

        return valid



from collections import defaultdict
import json
import os

import pandas as pd

from tuneconfig.experiment import Experiment


class Trial:

    def __init__(self, logdir, config, runs):
        self.logdir = logdir
        self.config = config
        self.runs = runs

    @property
    def results(self):
        return sorted(self[0])

    @property
    def metrics(self):
        return {
            result: sorted(df.columns) for result, df in self[0].items()
        }

    def info(self):
        print(f"<{self}>")
        print(f"ConfigIndex: {len(self.config)} parameters.")
        for param, value in self.config.items():
            print(f"  - {param} = {value}")
        print(f"RunIndex: {len(self)} runs.")
        print(f"ResultIndex: {len(self.results)} result files.")
        for result, df in self[0].items():
            print(f">> File '{result}' :")
            df.info()

    def describe(self):
        for result, df in self.stats().items():
            print(f">> Stats for '{result}' :")
            print(df)
            print()

    def stats(self):
        stats_ = defaultdict(list)

        for run, results in self.runs.items():
            for result, df in results.items():
                stats_[result].append(df)

        for result, data in stats_.items():
            data = pd.concat(data)
            stats_[result] = data.groupby(data.index, sort=False).agg([
                "min", "max", "mean", "std"
            ])

        return stats_

    @classmethod
    def from_directory(cls, dirname):
        # config
        with open(os.path.join(dirname, "config.json"), "r") as file:
            config = json.load(file)

        # runs
        runs = defaultdict(dict)
        for run_dir in Experiment.get_run_dirs(dirname):
            for path in os.listdir(run_dir):
                basename, extension = os.path.splitext(path)
                if extension == ".csv":
                    filepath = os.path.join(run_dir, path)
                    df = pd.read_csv(filepath)
                    runs[run_dir][basename] = df

        return Trial(dirname, config, runs)

    def __str__(self):
        return f"Trial(logdir='{self.logdir}'"

    def __len__(self):
        return len(self.runs)

    def __getitem__(self, i):
        return list(self.runs.items())[i][1]



import pytest

from tuneconfig.plotting import ExperimentPlotter


@pytest.fixture(scope="module")
def plotter(analysis):
    return ExperimentPlotter(analysis)


def test_plot(plotter):
    targets = ["data:foo", "data:baz"]
    anchors = ["batch=32", "lr=0.01"]
    x_axis, y_axis = None, "optimizer"
    kwargs = {
        "target_x_axis_label": "Epochs",
        "target_y_axis_label": "Loss"
    }

    plotter.plot(targets, anchors, x_axis, y_axis, **kwargs)



import pandas as pd
import pytest


def test_info(trial):
    trial.info()


def test_describe(trial):
    trial.describe()


def test_stats(trial):
    stats = trial.stats()
    assert isinstance(stats, dict)
    assert len(stats) == len(trial.metrics)

    for result, data in stats.items():
        assert result in trial.results
        assert isinstance(data, pd.DataFrame)



import pytest

from tuneconfig.trial import Trial


def test_setup(analysis):
    trials = analysis._trials
    assert isinstance(trials, dict)
    assert all(isinstance(trial, Trial) for trial in trials.values())
    assert all(hasattr(trial, "config") for trial in trials.values())
    assert all(hasattr(trial, "runs") for trial in trials.values())


def test_params(analysis):
    print(analysis.params)


def test_metrics(analysis):
    print(analysis.metrics)


def test_info(analysis):
    analysis.info()



import itertools
import multiprocessing as mp
import os
import pandas as pd
import re

from tqdm import tqdm


class Experiment:
    """
    Experiment -> trials -> runs.
    """

    CONFIG_FILE = "config.json"
    RUN_DIR_REGEX = r"run\d+$"

    def __init__(self, config_iterator, logdir, name=None, description=None):
        self.config_iterator = config_iterator
        self.logdir = logdir
        self.name = name
        self.description = description

    def start(self):
        config_files = self.config_iterator.dump(self.logdir)

    def run(self, exec_func, num_samples, num_workers=None, verbose=False):
        """
        Executes the trial runner function a given number of times.

        Args:
            exec_func (Callable[Dict] -> Result): The trial runner function.
            num_samples (int): The number of runs per trial.
            num_workers (int): The number of worker processes.
        """
        results = {}

        total_num_trials = len(self.config_iterator)

        for i, config in enumerate(self.config_iterator):
            if verbose:
                print()
                print(f">> Trial #{i+1}/{total_num_trials}")
                print(config)

            trial_id, trial_dir = self._get_trial(config)
            range_num_samples = self._get_run_ids(trial_dir, num_samples)

            trial_configs = []
            for i, j in enumerate(range_num_samples):
                trial_configs.append({
                    **config,
                    "run_id": i,
                    "logdir": os.path.join(trial_dir, f"run{j}"),
                })

            pool = mp.Pool(
                processes=num_workers,
                initializer=tqdm.set_lock,
                initargs=(tqdm.get_lock(),)
            )
            results[trial_id] = pool.map(exec_func, trial_configs)
            pool.close()
            pool.join()

        return results

    def _get_trial(self, config):
        trial_id = self.config_iterator._trial_id(config)
        trial_dir = os.path.join(self.logdir, trial_id)
        return trial_id, trial_dir

    @classmethod
    def get_run_dirs(cls, trial_dir):
        return [
            f.path for f in os.scandir(trial_dir)
            if re.search(cls.RUN_DIR_REGEX, f.path)
        ]

    @classmethod
    def _get_run_ids(cls, trial_dir, num_samples):
        previous_run_dirs = cls.get_run_dirs(trial_dir)
        start_id = len(previous_run_dirs)
        return range(start_id, start_id + num_samples)

    @classmethod
    def is_run_dir(cls, basename):
        return re.search(cls.RUN_DIR_REGEX, basename)

    @classmethod
    def is_trial_dir(cls, dirname):
        return cls.CONFIG_FILE in os.listdir(dirname)
